from __future__ import annotations

import warnings
from functools import partial
from pathlib import Path
from typing import Any, Callable, Iterable, Iterator, overload

import ase
import numpy as np
from ase import Atoms
from yaml import dump

from .database import DatabaseEntry, backend
from .utils import (
    LazyMapping,
    intersect,
    k_fold_split,
    random_split,
    split_keeping_ratio,
    union,
)


class AtomsDataset:
    """An immutable wrapper around a list of :class:`ase.Atoms` objects."""

    def __init__(self, structures: list[ase.Atoms]):
        if len(structures) == 1:
            warnings.warn(
                "Creating a dataset with a single structure. "
                "Typically, datasets contain multiple structures - "
                "did you mean to do this?",
                stacklevel=2,
            )

        self.structures = structures

        keys, loader = _get_info_loader(structures)
        self.info: LazyMapping[str, Any] = LazyMapping(keys, loader)
        """
        A key-value mapping of per-structure properties.
        
        Examples
        --------
        Scalar per-structure properties are turned into an 
        :class:`~numpy.ndarray`:

        >>> dataset = load_dataset("C-GAP-17")
        >>> dataset.info["energy"]
        array([-9847.661671, -9886.440245, -9916.681692, ...,  -632.464825,
                -632.146922,  -632.944571])

        Non-scalar properties are concatenated along a new axis:

        >>> c_gap_20 = load_dataset("C-GAP-20")
        >>> c_gap_20.info["virial"].shape
        (6088, 3, 3)
        """

        keys, loader = _get_arrays_loader(structures)
        self.arrays: LazyMapping[str, np.ndarray] = LazyMapping(keys, loader)
        """
        A key-value mapping of per-atom properties.
        
        Example
        -------
        Per-atom properties are concatenated along the first axis:

        >>> dataset = load_dataset("C-GAP-17")
        >>> dataset.arrays["forces"].shape
        (284965, 3)
        """

    @property
    def structure_sizes(self) -> np.ndarray:
        """The number of atoms in each structure."""
        return np.array([len(s) for s in self.structures])

    @property
    def n_atoms(self) -> int:
        """The total number of atoms in the dataset."""
        return int(self.structure_sizes.sum())

    def filter_by(
        self, *functions: Callable[[ase.Atoms], bool], **info_kwargs: Any
    ) -> AtomsDataset:
        """
        Return a new dataset containing only the structures that match the
        given criteria.

        Parameters
        ----------
        dataset
            The dataset to filter.
        functions
            Functions to filter the dataset by. Each function should take an
            ASE Atoms object as input and return a boolean.
        info_kwargs
            Keyword arguments to filter the dataset by. Only atoms objects with
            matching info keys and values will be returned.

        Example
        -------
        Get small, amorphous structures with large forces:

        .. code-block:: pycon
            :emphasize-lines: 3-7

            >>> from load_atoms import load_dataset
            >>> dataset = load_dataset("C-GAP-17")
            >>> dataset.filter_by(
            ...     lambda structure: len(structure) < 50,
            ...     lambda structure: structure.arrays["force"].max() > 5,
            ...     config_type="bulk_amo"
            ... )
            Dataset:
                structures: 609
                atoms: 23,169
                species:
                    C: 100.00%
                properties:
                    per atom: (force)
                    per structure: (config_type, detailed_ct, split, energy)
        """

        def matches_info(structure: ase.Atoms) -> bool:
            for key, value in info_kwargs.items():
                if structure.info.get(key, None) != value:
                    return False
            return True

        functions = (*functions, matches_info)

        def the_filter(structure: ase.Atoms) -> bool:
            return all(f(structure) for f in functions)

        return AtomsDataset(list(filter(the_filter, self)))

    def random_split(
        self,
        splits: list[float] | list[int],
        seed: int = 42,
        keep_ratio: str | None = None,
    ) -> list[AtomsDataset]:
        r"""
        Randomly split the dataset into multiple, disjoint parts.

        Parameters
        ----------
        splits
            The number of structures to put in each split.
            If a list of :class:`float`\ s, the splits will be
            calculated as a fraction of the dataset size.
        seed
            The random seed to use for shuffling the dataset.
        keep_ratio
            If not :code:`None`, splits will be generated to maintain the
            ratio of structures in each split with the specified :code:`.info`
            value.

        Returns
        -------
        list[AtomsDataset]
            A list of new datasets, each containing a subset of the original

        Examples
        --------
        Split a :code:`dataset` into 80% training and 20% test sets:

        >>> train, test = dataset.random_split([0.8, 0.2])

        Split a :code:`dataset` into 3 parts:

        >>> train, val, test = dataset.random_split([1_000, 100, 100])

        Maintain the ratio of :code:`config_type` values in each split:

        .. code-block:: pycon
            :emphasize-lines: 16-19

            >>> from load_atoms import load_dataset
            >>> import numpy as np
            >>> # helper function
            >>> def ratios(thing):
            ...     values, counts = np.unique(thing, return_counts=True)
            ...     max_len = max(len(str(v)) for v in values)
            ...     for v, c in zip(values, counts / counts.sum()):
            ...         print(f"{v:>{max_len}}: {c:>6.2%}")
            ...
            >>> dataset = load_dataset("C-GAP-17")
            >>> ratios(dataset.info["config_type"])
              bulk_amo: 75.28%
            bulk_cryst:  8.83%
                 dimer:  0.66%
              surf_amo: 15.23%
            >>> train, val, test = dataset.random_split(
            ...     [0.6, 0.2, 0.2],
            ...     keep_ratio="config_type"
            ... )
            >>> ratios(train.info["config_type"])
              bulk_amo: 75.28%
            bulk_cryst:  8.83%
                 dimer:  0.66%
              surf_amo: 15.23%
        """

        if keep_ratio is None:
            return [
                AtomsDataset(split_structures)
                for split_structures in random_split(
                    self.structures, splits, seed
                )
            ]

        if keep_ratio not in self.info:
            raise KeyError(
                f"Unknown key {keep_ratio}. "
                "Available keys are: " + ", ".join(self.info.keys)
            )

        if isinstance(splits[0], int):  # TODO: fix type error
            final_sizes: list[int] = splits  # type: ignore
        else:
            final_sizes = [int(s * len(self)) for s in splits]  # type: ignore

        normalised_fractional_splits = [s / sum(splits) for s in splits]

        structure_splits: list[list[Atoms]] = split_keeping_ratio(
            self.structures,
            group_ids=self.info[keep_ratio],
            splitting_function=partial(
                random_split, seed=seed, splits=normalised_fractional_splits
            ),
        )

        def choose_n(structures: list[Atoms], n: int) -> list[Atoms]:
            idxs = np.random.RandomState(seed).permutation(len(structures))
            return [structures[i] for i in idxs[:n]]

        return [
            AtomsDataset(choose_n(split, size))
            for split, size in zip(structure_splits, final_sizes)
        ]

    def k_fold_split(
        self,
        k: int = 5,
        fold: int = 0,
        shuffle: bool = True,
        seed: int = 42,
        keep_ratio: str | None = None,
    ) -> tuple[AtomsDataset, AtomsDataset]:
        """
        Generate (an optionally shuffled) train/test split for cross-validation.

        Parameters
        ----------
        k
            The number of folds to use.
        fold
            The fold to use for testing.
        shuffle
            Whether to shuffle the dataset before splitting.
        seed
            The random seed to use for shuffling the dataset.
        keep_ratio
            If not :code:`None`, splits will be generated to maintain the
            ratio of structures in each split with the specified :code:`.info`
            value.

        Returns
        -------
        Tuple[Dataset, Dataset]
            The train and test datasets.

        Example
        -------

        Basic usage:

        .. code-block:: pycon
            :emphasize-lines: 2

            >>> for i in range(5):
            ...     train, test = dataset.k_fold_split(k=5, fold=i)
            ...     ...  # do something, e.g. train a model

        Maintain the ratio of :code:`config_type` values in each split
        (see also :func:`~load_atoms.AtomsDataset.random_split` for a more
        detailed example of this feature):

        .. code-block:: pycon

            >>> train, test = dataset.k_fold_split(
            ...     k=5, fold=0, keep_ratio="config_type"
            ... )

        """

        if k < 2:
            raise ValueError("k must be at least 2")

        fold = fold % k

        if shuffle:
            idxs = np.random.RandomState(seed).permutation(len(self))
        else:
            idxs = np.arange(len(self))

        if keep_ratio is None:
            train_idxs, test_idxs = k_fold_split(idxs.tolist(), k, fold)
        else:
            if keep_ratio not in self.info:
                raise KeyError(
                    f"Unknown key {keep_ratio}. "
                    "Available keys are: " + ", ".join(self.info.keys)
                )
            if not shuffle:
                raise ValueError(
                    "Keep ratio splits are only supported when shuffling."
                )

            group_ids = self.info[keep_ratio][idxs]
            train_idxs, test_idxs = split_keeping_ratio(
                idxs.tolist(), group_ids, partial(k_fold_split, k=k, fold=fold)
            )

        return self[train_idxs], self[test_idxs]

    def __len__(self) -> int:
        """The number of structures in the dataset."""
        return len(self.structures)

    @overload
    def __getitem__(self, index: int) -> ase.Atoms:
        ...

    @overload
    def __getitem__(self, index: slice) -> AtomsDataset:
        ...

    @overload
    def __getitem__(self, index: np.ndarray) -> AtomsDataset:
        ...

    @overload
    def __getitem__(self, index: Iterable[int]) -> AtomsDataset:
        ...

    def __getitem__(self, index: Any):
        """
        Index the dataset.

        Parameters
        ----------
        index
            specifies the structure/s to select from the dataset.

        Examples
        --------

        >>> from load_atoms import load_dataset
        >>> dataset = load_dataset("C-GAP-17")
        The C-GAP-17 dataset is covered by the CC BY-NC-SA 4.0 license.
        Please cite the C-GAP-17 dataset if you use it in your work.

        Get the first structure in the dataset:

        >>> dataset[0]
        Atoms(symbols='C64', pbc=True, cell=[9.483, 9.483, 9.483], force=...)

        Create a new dataset with the first 10 structures:

        >>> dataset[:10]
        Dataset:
            structures: 10
            atoms: 640
            species:
                C: 100.00%
            properties:
                per atom: (force)
                per structure: (config_type, detailed_ct, split, energy)

        Create a new dataset of high energy structures using a mask
        (see :func:`~load_atoms.AtomsDataset.filter_by`
        for a more convenient way to do this):

        >>> mask = dataset.info["energy"] / dataset.structure_sizes > -155
        >>> dataset[mask]
        Dataset:
            structures: 36
            atoms: 444
            species:
                C: 100.00%
            properties:
                per atom: (force)
                per structure: (config_type, detailed_ct, split, energy)
        """
        # if the passed index is a slice, return a new Dataset object:
        if isinstance(index, slice):
            return AtomsDataset(self.structures[index])

        # if the index is iterable, return a new Dataset object:
        if hasattr(index, "__iter__"):
            # if the index is a np index, we want to keep the same behaviour
            # (e.g. passing array of indices, or a boolean array)
            if isinstance(index, np.ndarray):
                to_keep = np.arange(len(self))[index]
                return AtomsDataset([self.structures[i] for i in to_keep])

            # some other iterable, e.g. a list of indices
            return AtomsDataset([self.structures[i] for i in index])  # type: ignore

        # otherwise, we assume the index is an integer,
        # and return a single structure
        return self.structures[int(index)]

    def __iter__(self) -> Iterator[ase.Atoms]:
        """
        Iterate over the structures in the dataset.
        """

        return iter(self.structures)

    def __repr__(self):
        return summarise_dataset(self.structures)


class DescribedDataset(AtomsDataset):
    def __init__(
        self,
        structures: list[Atoms],
        description: DatabaseEntry,
    ):
        super().__init__(structures)
        self.description = description

    @classmethod
    def from_id(
        cls,
        dataset_id: str,
        root: Path | (str | None) = None,
    ) -> AtomsDataset:
        """
        Load a dataset by id.

        Parameters
        ----------
        dataset_id : str
            the id of the dataset to load
        root : Union[Path, str, None], optional
            the root directory to cache the dataset to, by default None
        verbose : bool, optional
            whether to print information about the dataset, by default True
        """

        if root is None:
            root = Path.home() / ".load-atoms"
        root = Path(root)

        all_structures, info = backend.load_structures(dataset_id, root)

        # remove annoying automatic ASE calculators
        for structure in all_structures:
            structure.calc = None

        return cls(all_structures, info)

    def __repr__(self):
        return summarise_dataset(self.structures, self.description)


def _get_info_loader(
    structures: list[Atoms],
) -> tuple[list[str], Callable[[str], Any]]:
    keys = intersect(structure.info.keys() for structure in structures)

    def loader(key: str):
        return np.array([structure.info[key] for structure in structures])

    return list(keys), loader


def _get_arrays_loader(
    structures: list[Atoms],
) -> tuple[list[str], Callable[[str], Any]]:
    keys = intersect(structure.arrays.keys() for structure in structures)

    def loader(key: str):
        return np.concatenate(
            [structure.arrays[key] for structure in structures]
        )

    return list(keys), loader


def summarise_dataset(
    structures: list[Atoms] | AtomsDataset,
    description: DatabaseEntry | None = None,
) -> str:
    name = description.name if description is not None else "Dataset"
    N = len(structures)
    number_atoms = sum([len(structure) for structure in structures])

    per_atom_properties = sorted(
        intersect(structure.arrays.keys() for structure in structures)
        - {"numbers", "positions"}
    )
    per_structure_properties = sorted(
        intersect(structure.info.keys() for structure in structures)
    )

    species = sorted(
        union(structure.get_chemical_symbols() for structure in structures)
    )
    species_counts = {
        s: sum(
            [
                structure.get_chemical_symbols().count(s)
                for structure in structures
            ]
        )
        for s in species
    }
    species_counts = {
        k: f"{v / number_atoms:.2%}"
        for k, v in sorted(
            species_counts.items(), key=lambda item: item[1], reverse=True
        )
    }

    if N >= 1000:
        N = f"{N:,}"
    if number_atoms >= 1000:
        number_atoms = f"{number_atoms:,}"

    fields = {
        "structures": N,
        "atoms": number_atoms,
        "species": species_counts,
        "properties": {
            "per atom": "(" + ", ".join(per_atom_properties) + ")",
            "per structure": "(" + ", ".join(per_structure_properties) + ")",
        },
    }

    return dump({name: fields}, sort_keys=False, indent=4)



from __future__ import annotations

from pathlib import Path

import ase
from ase.io import read

from .atoms_dataset import AtomsDataset, DescribedDataset
from .visualisation import view

__version__ = "0.2.7"
__all__ = ["load_dataset", "view"]


def load_dataset(
    thing: str | list[ase.Atoms] | Path,
    root: str | Path | None = None,
) -> AtomsDataset:
    """
    Load a dataset by name or from a list of structures.

    Parameters
    ----------
    thing
        A dataset id, a list of structures, or a path to a file.
    root
        The root directory to use when loading a dataset by id. If not
        provided, the default root directory (:code:`~/.load-atoms`)
        will be used.

    Examples
    --------
    Load a dataset by id:

    >>> from load_atoms import load_dataset
    >>> dataset = load_dataset("QM9")
    ╭───────────────────────────────── QM9 ─────────────────────────────────╮
    │                                                                       │
    │   Downloading dsgdb9nsd.xyz.tar.bz2 ━━━━━━━━━━━━━━━━━━━━ 100% 00:09   │
    │   Extracting dsgdb9nsd.xyz.tar.bz2  ━━━━━━━━━━━━━━━━━━━━ 100% 00:18   │
    │   Processing files                  ━━━━━━━━━━━━━━━━━━━━ 100% 00:19   │
    │   Caching to disk                   ━━━━━━━━━━━━━━━━━━━━ 100% 00:02   │
    │                                                                       │
    │            The QM9 dataset is covered by the CC0 license.             │
    │        Please cite the QM9 dataset if you use it in your work.        │
    │          For more information about the QM9 dataset, visit:           │
    │                            load-atoms/QM9                             │
    ╰───────────────────────────────────────────────────────────────────────╯
    >>> dataset
    QM9:
        structures: 133,885
        atoms: 2,407,753
        species:
            H: 51.09%
            C: 35.16%
            O: 7.81%
            N: 5.80%
            F: 0.14%
        properties:
            per atom: (partial_charges)
            per structure: (
                A, B, C, Cv, G, H, U, U0, alpha,
                frequencies, gap, geometry, homo, inchi, index,
                lumo, mu, r2, smiles, zpve
            )


    Optionally save a dataset to an explicit root directory:

    >>> load_dataset("QM9", root="./my-datasets")

    Wrap a list of structures in a dataset:

    >>> load_dataset([Atoms("H2O"), Atoms("H2O2")])

    Load a dataset from a file:

    >>> load_dataset("path/to/file.xyz")
    """

    if isinstance(thing, list) and all(isinstance(s, ase.Atoms) for s in thing):
        # thing is a list of structures
        return AtomsDataset(thing)

    if not isinstance(thing, (Path, str)):
        raise TypeError(
            f"Could not load dataset from {thing}. "
            "Please provide a string, a list of structures, "
            "or a path to a file."
        )

    if Path(thing).exists():
        # thing is a string/path to a file that exists
        # assume it is a file containing structures and load them
        structures = read(Path(thing), index=":")
        return AtomsDataset(structures)  # type: ignore

    if isinstance(thing, Path):
        # thing is a path to a file that does not exist
        raise ValueError(f"The provided path does not exist. ({thing})")

    # assume thing is a dataset ID, and try to load it
    return DescribedDataset.from_id(thing, root)



from __future__ import annotations

from pathlib import Path

import requests

from load_atoms.progress import Progress


def download(url: str, local_path: Path, progress: Progress):
    """
    Download a file from the given url to the given path.

    If path is a file, the file will be downloaded to that path.
    Else, the file will be downloaded to the given path, with the same name as
    the file at the given url.

    Parameters
    ----------
    url
        The url to download the file from.
    local_path
        The path to download the file to.
    progress
        The progress bar to add the download to.
    """

    if local_path.is_dir():
        local_path = local_path / Path(url).name

    with requests.get(url, stream=True) as response, progress.new_task(
        f"Downloading {local_path.name}"
    ) as task, open(local_path, "wb") as f:
        response.raise_for_status()
        file_size = int(response.headers["content-length"])
        task.update(total=file_size)
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                task.update(advance=len(chunk))



# explicitly not using future annotations since this is not supported
# by pydantic for python versions we want to target
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import yaml
from ase import Atoms
from pydantic import BaseModel, field_validator

from load_atoms.progress import Progress
from load_atoms.utils import BASE_REMOTE_URL, valid_checksum

from .processing import default_processing, parse_steps

valid_licenses = {
    "CC BY-NC-SA 4.0": "https://creativecommons.org/licenses/by-nc-sa/4.0/deed.en",
    "CC BY-NC 4.0": "https://creativecommons.org/licenses/by-nc/4.0/deed.en",
    "CC BY 4.0": "https://creativecommons.org/licenses/by/4.0/deed.en",
    "CC0": "https://creativecommons.org/publicdomain/zero/1.0/",
    "MIT": "https://opensource.org/licenses/MIT",
    "GPLv3": "https://www.gnu.org/licenses/gpl-3.0.html",
}

valid_categories = ["Benchmarks", "Potential Fitting", "Synthetic Data"]


class PropertyDescription(BaseModel):
    """
    Holds a description of a property, such that it can be automatically
    validated upon creation.
    """

    desc: str
    """A description of the property"""

    units: Optional[str] = None
    """The units of the property"""


class FileInformation(BaseModel):
    """
    Holds metadata for a remote file

    Partial speficiation are allowed:

    Speficiations of the following form assume that the file is available at
    :code:`database/{name}/{filename}`.

    .. code-block:: yaml

        name: <filename>
        hash: <checksum>

    Speficiations of the following form assume that the :code:`filename`
    is the last part of the URL.

    .. code-block:: yaml

        url: <url>
        hash: <checksum>
    """

    name: str
    hash: str
    url: str

    @field_validator("hash")
    def validate_hash(cls, v):
        if not valid_checksum(v):
            raise ValueError(f"Invalid checksum: {v}")
        return v


class DatabaseEntry(BaseModel):
    """
    Holds all the required metadata for a named dataset, such that it can be
    automatically downloaded using
    :func:`load_dataset(dataset_name) <load_atoms.load_dataset>`, and so that
    documentation can be automatically generated.

    Subclassing pydantic's BaseModel, which means that automatic
    validation occurs upon creation.
    """

    name: str
    """The name of the dataset"""

    year: int
    """The year the dataset was created"""

    description: str
    """A description of the dataset (in :code:`.rst` format)"""

    files: List[FileInformation]
    """
    A list of files that are part of the dataset, along with their checksums
    and URLs.
    """

    category: str
    """
    The category of the dataset 
    (e.g. :code:`"Potential Fitting"`, :code:`"Benchmarks"`)
    """

    citation: Optional[str] = None
    """A citation for the dataset (in BibTeX format)"""

    license: Optional[str] = None
    """The license identifier of the dataset (e.g. :code:`"CC BY-NC-SA 4.0"`)"""

    representative_structure: Optional[int] = None
    """The index of a representative structure (for visualisation purposes)"""

    per_atom_properties: Optional[Dict[str, PropertyDescription]] = None
    """A mapping from per-atom properties to their descriptions"""

    per_structure_properties: Optional[Dict[str, PropertyDescription]] = None
    """A mapping from per-structure properties to their descriptions"""

    processing: Callable[[Path, Progress], List[Atoms]]
    """
    A function to turn the root downloaded path into the dataset's structures
    """

    @field_validator("category")
    def validate_category(cls, v):
        if v not in valid_categories:
            raise ValueError(
                f"Invalid category: {v}. Must be one of {valid_categories}"
            )
        return v

    @field_validator("license")
    def validate_license(cls, v):
        if v not in valid_licenses:
            raise ValueError(
                f"Invalid license: {v}. Must be one of {list(valid_licenses)}"
            )
        return v

    @field_validator("citation")
    def validate_citation(cls, v):
        v = v.strip()
        if v.startswith("@") and v.endswith("}"):
            return v
        raise ValueError(f"Invalid BibTeX: {v}")

    @classmethod
    def from_yaml_file(cls, path: Union[Path, str]) -> "DatabaseEntry":
        with open(path) as f:
            data = yaml.safe_load(f)

        # process defaults for file structure
        for file in data["files"]:
            if "url" not in file:
                file["url"] = BASE_REMOTE_URL + f"{data['name']}/{file['name']}"
            elif "name" not in file:
                file["name"] = file["url"].split("/")[-1]

        # parse processing : TODO: pydantic-ify this
        if "processing" in data:
            data["processing"] = parse_steps(data["processing"])
        else:
            files = data["files"]
            if not len(files) == 1:
                raise ValueError(
                    "If no processing is provided, the dataset must have "
                    "exactly one file."
                )
            data["processing"] = default_processing(files[0]["name"])

        try:
            return cls(**data)
        except Exception as e:
            raise ValueError(
                f"Error loading dataset description from {path}"
            ) from e

    @classmethod
    def remote_url_for(cls, dataset_id):
        return BASE_REMOTE_URL + f"{dataset_id}/{dataset_id}.yaml"



from __future__ import annotations

from pathlib import Path

from ase import Atoms
from ase.io import read

from load_atoms.database.processing import UnZip


def process(root: Path) -> list[Atoms]:
    """
    Read out the AC-2D-22 dataset from the downloaded Zenodo archive.
    """
    root = UnZip()(root)
    root = root / "data/fig_3"

    def read_structures(archive_dir) -> list[Atoms]:
        extracted = UnZip(file="fin_run.tar.gz")(archive_dir)
        return [read(file) for file in sorted(extracted.glob("**/*.xyz"))]  # type: ignore

    table = [
        (
            "loc_run_fig_3a/Final_structures",
            {
                "config": "continuous random network",
                "beta": 2.0,
                "criterion": "local energy",
            },
        ),
        (
            "NN_run_fig_3b/Final_structures_pristine",
            {
                "config": "paracrystalline",
                "beta": 2.0,
                "criterion": "nearest-neighbour energy",
            },
        ),
        (
            "NN_run_fig_3b/Final_structures_thermal",
            {
                "config": "paracrystalline",
                "beta": 2.0,
                "criterion": "nearest-neighbour energy",
            },
        ),
        (
            "loc_run_fig_3c/beta_50/Final_structures",
            {
                "config": "paracrystalline",
                "beta": 50.0,
                "criterion": "local energy",
            },
        ),
        (
            "loc_run_fig_3c/beta_60/Final_structures",
            {
                "config": "paracrystalline",
                "beta": 60.0,
                "criterion": "local energy",
            },
        ),
        (
            "fig_3d/total_beta_0.8/Final_structures",
            {
                "config": "paracrystalline",
                "beta": 0.8,
                "criterion": "total energy",
            },
        ),
    ]

    all_structures: list[Atoms] = []

    for path, info in table:
        structures = read_structures(root / path)
        for structure in structures:
            structure.info = info
        all_structures.extend(structures)

    renaming = {
        "Forces": "forces",
        "Energy_per_atom": "local_energy",
        "NN_Energy_per_atom": "nn_local_energy",
    }
    drop = ["c_1"]

    for structure in all_structures:
        for old, new in renaming.items():
            structure.arrays[new] = structure.arrays.pop(old)
        for key in drop:
            del structure.arrays[key]
        # rotate 90 degrees around the x-axis
        structure.rotate(90, "x", center="COU", rotate_cell=True)

    return all_structures



from pathlib import Path

import numpy as np
import pytest
from ase import Atoms
from ase.io import read, write
from load_atoms import load_dataset
from load_atoms.atoms_dataset import AtomsDataset, summarise_dataset
from load_atoms.utils import UnknownDatasetException
from setup import TESTING_DIR

STRUCTURES = [Atoms("H2O"), Atoms("H2O2")]
GAP17 = load_dataset("C-GAP-17", root=TESTING_DIR)


def _is_water_dataset(dataset):
    assert len(dataset) == 2
    assert set(dataset[0].symbols) == {"H", "O"}


def test_dataset_from_structures():
    dataset = load_dataset(STRUCTURES)
    _is_water_dataset(dataset)


def test_dataset_writeable_and_readable(tmp_path):
    dataset = load_dataset(STRUCTURES)
    write(tmp_path / "test.xyz", dataset)  # type: ignore

    dataset2 = load_dataset(read(tmp_path / "test.xyz", index=":"))  # type: ignore
    _is_water_dataset(dataset2)

    dataset3 = load_dataset(tmp_path / "test.xyz")
    _is_water_dataset(dataset3)


@pytest.mark.filterwarnings("ignore:Creating a dataset with a single structure")
def test_indexing():
    dataset = load_dataset(STRUCTURES)

    structure = dataset[0]
    assert isinstance(
        structure, Atoms
    ), "Indexing should return an Atoms object"
    assert (
        structure is STRUCTURES[0]
    ), "Indexing should return the same structure"

    sub_dataset = dataset[1:]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Slicing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Slicing should return the correct number of structures"

    indices = np.array([0, 1, 0, 1])
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 4
    ), "Indexing should return the correct number of structures"

    indices = np.array([True, False])
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Indexing should return the correct number of structures"

    indices = [0, 1, 1]
    sub_dataset = dataset[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 3
    ), "Indexing should return the correct number of structures"


def test_can_load_from_id(tmp_path):
    assert len(GAP17) == 4530

    with pytest.raises(UnknownDatasetException):
        load_dataset("made_up_dataset", root=tmp_path)


def test_summarise():
    dataset = load_dataset(STRUCTURES)
    summary = summarise_dataset(dataset)
    assert "Dataset" in summary, "The summary should contain the dataset name"

    summary = repr(GAP17)
    assert "energy" in summary, "The summary should contain the property names"


def test_useful_error_message():
    with pytest.raises(
        TypeError,
        match="provide a string, a list of structures, or a path to a file.",
    ):
        load_dataset(1)  # type: ignore

    with pytest.raises(
        ValueError,
        match="The provided path does not exist.",
    ):
        load_dataset(Path("made_up_file.xyz"))


def test_useful_warning(tmp_path):
    structure = Atoms("H2O")
    write(tmp_path / "test.xyz", structure)  # type: ignore

    with pytest.warns(
        UserWarning,
        match="single structure",
    ):
        load_dataset(tmp_path / "test.xyz")


def test_info_and_arrays():
    assert "energy" in GAP17.info
    assert isinstance(GAP17.info["energy"], np.ndarray)

    assert "positions" in GAP17.arrays
    assert GAP17.arrays["positions"].shape[-1] == 3


def test_properties():
    dataset = load_dataset(STRUCTURES)

    assert len(dataset) == 2
    assert dataset.n_atoms == 7
    assert (dataset.structure_sizes == [3, 4]).all()


def _get_proportion(config_type, dataset):
    return np.sum(dataset.info["config_type"] == config_type) / len(dataset)


def test_random_split():
    # request integer splits
    train, test = GAP17.random_split([100, 50])
    assert len(train) == 100
    assert len(test) == 50

    # request float splits
    n = len(GAP17)
    a, b, c = GAP17.random_split([0.5, 0.25, 0.25])
    assert len(a) == n // 2
    assert len(b) in [n // 4, n // 4 + 1]
    assert len(c) in [n // 4, n // 4 + 1]

    # test keep ratio
    assert "config_type" in GAP17.info
    assert "bulk_amo" in GAP17.info["config_type"]

    # ... with floats
    a, b = GAP17.random_split([0.5, 0.5], keep_ratio="config_type")
    assert len(a) + len(b) == len(GAP17)
    assert np.isclose(
        _get_proportion("bulk_amo", a), _get_proportion("bulk_amo", GAP17)
    )

    # ... with ints
    a, b = GAP17.random_split([500, 500], keep_ratio="config_type")
    assert np.isclose(
        _get_proportion("bulk_amo", a),
        _get_proportion("bulk_amo", GAP17),
        atol=0.01,
    )
    assert len(a) == len(b) == 500

    # test error
    assert "made_up" not in GAP17.info
    with pytest.raises(KeyError):
        GAP17.random_split([0.5, 0.5], keep_ratio="made_up")


def test_k_fold_split():
    # error messages
    with pytest.raises(ValueError, match="k must be at least 2"):
        GAP17.k_fold_split(k=1)

    # atoms are not hashable, so we set a unique id for each structure
    for i, structure in enumerate(GAP17):
        structure.info["id"] = i

    a, b = GAP17.k_fold_split(5, fold=0, shuffle=False)
    assert a[0] == GAP17[0]

    a, b = GAP17.k_fold_split(5, fold=0)
    assert len(a) == 3624
    assert len(b) == 906
    assert set(a.info["id"]) & set(b.info["id"]) == set()

    c, d = GAP17.k_fold_split(5, fold=1)
    # ensure that b is completely wihin c
    assert set(b.info["id"]) <= set(c.info["id"])

    # ensure that the folds completely cover the dataset
    all_test = []
    for i in range(5):
        _, test = GAP17.k_fold_split(5, fold=i)
        all_test.extend(test.info["id"])

    assert len(set(all_test)) == len(GAP17)

    # test keep ratio
    a, b = GAP17.k_fold_split(k=5, fold=0, keep_ratio="config_type")
    assert np.isclose(
        _get_proportion("bulk_amo", a), _get_proportion("bulk_amo", GAP17)
    )

    # test error
    assert "made_up" not in GAP17.info
    with pytest.raises(KeyError):
        GAP17.k_fold_split(5, fold=0, keep_ratio="made_up")

    with pytest.raises(ValueError, match="only supported when shuffling"):
        GAP17.k_fold_split(5, fold=0, keep_ratio="config_type", shuffle=False)



from __future__ import annotations

import hashlib
import string
from collections import defaultdict
from pathlib import Path
from typing import Callable, Generic, Iterable, Sequence, TypeVar

import numpy as np

FRONTEND_URL = "https://jla-gardner.github.io/load-atoms/datasets/"
BASE_REMOTE_URL = "https://github.com/jla-gardner/load-atoms/raw/main/database/"


def generate_checksum(file_path: Path | str) -> str:
    """Generate a checksum for a file."""

    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()[:12]


def valid_checksum(hash: str) -> bool:
    """Check if a hash is valid."""
    if len(hash) != 12:
        return False
    return all(c in string.hexdigits for c in hash)


def matches_checksum(file_path: Path, hash: str) -> bool:
    """Check if a file matches a given hash."""
    return generate_checksum(file_path) == hash


T = TypeVar("T")
Y = TypeVar("Y")
G = TypeVar("G")


class LazyMapping(Generic[T, Y]):
    """
    A mapping that lazily loads its values.

    Concretely, the first time a key is accessed, the loader function is called
    to get the value for that key. Subsequent accesses to the same key will
    return the same value without calling the loader function again.

    Parameters
    ----------
    keys: Sequence[T]
        The keys of the mapping.
    loader: Callable[[T], Y]
        A function that takes a key and returns a value.

    Examples
    --------

    >>> def loader(key):
    ...     print(f"Loading value for key={key}")
    ...     return key * 2
    ...
    >>> mapping = LazyMapping([1, 2, 3], loader)
    >>> mapping[3]
    Loading value for key=3
    6
    >>> mapping[3]
    6
    >>> 1 in mapping
    True
    >>> 4 in mapping
    False
    """

    def __init__(
        self,
        keys: Sequence[T],
        loader: Callable[[T], Y],
    ):
        self.keys = keys
        self.loader = loader
        self._mapping = {}

    def __getitem__(self, key: T) -> Y:
        if key not in self.keys:
            raise KeyError(key)
        if key not in self._mapping:
            self._mapping[key] = self.loader(key)
        return self._mapping[key]

    def __contains__(self, key: T):
        return key in self.keys

    def __repr__(self) -> str:
        return f"LazyMapping(keys={self.keys})"


def frontend_url(dataset_info):
    """Get the URL for a dataset's information page."""
    return FRONTEND_URL + dataset_info.name + ".html"


class UnknownDatasetException(Exception):
    def __init__(self, dataset_id):
        super().__init__(f"Unknown dataset: {dataset_id}")


def union(things: Iterable[Iterable]):
    """Get the set union of a list of iterables."""
    return set.union(*map(set, things), set())


def intersect(things: Iterable[Iterable]):
    """Get the set intersection of a list of iterables."""
    sets = list(map(set, things))
    if not sets:
        return set()
    return set.intersection(*sets)


def lpad(thing: str, length: int = 4, fill: str = " "):
    """Left pad a string with a given fill character."""
    sep = f"{fill * length}"
    return sep + thing.replace("\n", f"\n{sep}")


def random_split(
    things: list[T],
    splits: list[int] | list[float],
    seed: int = 0,
) -> list[list[T]]:
    """Split a list into random chunks of given sizes."""

    if isinstance(splits[0], float):
        splits = [int(s * len(things)) for s in splits]
    if sum(splits) > len(things):
        raise ValueError(
            "The sum of the splits cannot exceed the dataset size."
        )

    cumulative_sum = np.cumsum(splits)

    idxs = np.random.RandomState(seed).permutation(len(things))
    return [
        [things[x] for x in idxs[i:j]]
        for i, j in zip([0, *cumulative_sum], cumulative_sum)
    ]


def k_fold_split(
    things: list[T],
    k: int,
    fold: int,
) -> tuple[list[T], list[T]]:
    assert 0 <= fold < k

    idxs = np.arange(len(things))
    idxs = np.roll(idxs, fold * len(things) // k)
    n_test = len(things) // k
    train, test = idxs[:-n_test], idxs[-n_test:]
    return [things[i] for i in train], [things[i] for i in test]


def split_keeping_ratio(
    things_to_split: list[T],
    group_ids: list[G],
    splitting_function: Callable[[list[T]], Sequence[list[T]]],
):
    assert len(things_to_split) == len(group_ids)

    # 1. separate into groups
    groups: dict[G, list[T]] = defaultdict(list)
    for thing, group_id in zip(things_to_split, group_ids):
        groups[group_id].append(thing)

    # 2. split each group
    splits_by_group: dict[G, Sequence[list[T]]] = {
        key: splitting_function(value) for key, value in groups.items()
    }

    # 3. merge the splits, thus keeping the ratio
    n_splits = len(splits_by_group[list(groups.keys())[0]])
    final_splits: list[list[T]] = [[] for _ in range(n_splits)]
    for group_splits in splits_by_group.values():
        for i, split in enumerate(group_splits):
            final_splits[i].extend(split)

    return final_splits



from __future__ import annotations

from datetime import timedelta

from rich.align import Align
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    ProgressColumn,
    SpinnerColumn,
    TaskID,
    TextColumn,
)
from rich.progress import Progress as RichProgress
from rich.table import Column, Table
from rich.text import Text


class TimeElapsedColumn(ProgressColumn):
    """Renders time elapsed."""

    def render(self, task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("--:--", style="black")
        delta = timedelta(seconds=int(elapsed))
        return Text(":".join(str(delta).split(":")[1:]), style="black")


class PercentColumn(ProgressColumn):
    """Renders percentage complete."""

    def render(self, task) -> Text:
        """Show percentage complete."""
        if task.completed is True:
            return Text("100%", style="black")
        if task.total is None:
            return Text("    ")
        return Text(f"{task.percentage:>3.0f}%", style="black")


class Progress:
    def __init__(self, description: str, transient: bool = False):
        self._progress = RichProgress(
            SpinnerColumn(
                spinner_name="point",
                finished_text="[bold green] ✓ [/bold green]",
                speed=0.5,
                style="black",
            ),
            TextColumn(
                "[progress.description]{task.description}",
                table_column=Column(max_width=40, overflow="fold"),
            ),
            BarColumn(bar_width=20),
            PercentColumn(),
            TimeElapsedColumn(),
        )
        self._table = Table.grid()
        self._table.add_row()
        self._table.add_row(Align(self._progress, align="center"))
        self._live = Live(
            Panel.fit(
                self._table,
                title=f"[bold]{description}",
            ),
            refresh_per_second=10,
            transient=transient,
        )

    def new_task(
        self,
        description: str,
        transient: bool = False,
        total: int | float | None = None,
        **kwargs,
    ) -> Task:
        return Task(
            self._progress.add_task(description, total=total, **kwargs),
            self._progress,
            transient,
        )

    def log_below(self, text: str, align: str = "center"):
        self._table.add_row(Align(Text.from_markup(text), align=align))  # type: ignore

    def __enter__(self):
        self._live.__enter__()
        return self

    def __exit__(self, *args):
        self._progress.refresh()
        self._live.refresh()
        self._live.__exit__(*args)


class Task:
    def __init__(self, task: TaskID, progress: RichProgress, transient: bool):
        self._task = task
        self._progress = progress
        self._transient = transient

    def update(self, **kwargs):
        self._progress.update(self._task, **kwargs)

    def complete(self, remove: bool = False):
        self._progress.update(self._task, completed=True, total=1)
        if remove:
            self._progress.remove_task(self._task)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if args[0] is None:
            self.complete(remove=self._transient)



from __future__ import annotations

import contextlib
import importlib
import shutil
from pathlib import Path
from typing import Generic, Protocol, TypeVar

import yaml
from ase import Atoms
from ase.io import read as ase_read

from load_atoms.progress import Progress

X, Y = TypeVar("X", contravariant=True), TypeVar("Y", covariant=True)


class Step(Protocol, Generic[X, Y]):
    """
    Represents a single step in a processing
    :class:`~load_atoms.database.processing.Chain`.
    """

    def __call__(self, x: X, progress: Progress | None = None) -> Y:
        """Perform the step, transforming the input into the output."""
        ...


class Chain(Step[X, Y]):
    """
    Represents a chain of processing steps.

    Parameters
    ----------
    steps
        The steps to perform in order.
    """

    def __init__(self, steps: list[Step]):
        self.steps = steps

    def __call__(self, x: X, progress: Progress | None = None) -> Y:
        """Perform the steps in order."""
        for step in self.steps:
            x = step(x, progress)
        return x  # type: ignore

    def __repr__(self):
        return f"Chain({', '.join(str(step) for step in self.steps)})"


_common_steps: dict[str, type[Step]] = dict()


def register_step(step: type[Step]):
    _common_steps[step.__name__] = step
    return step


def get_step_type(name: str) -> type[Step]:
    try:
        return _common_steps[name]
    except KeyError:
        raise ValueError(
            f"Unknown step: {name}. " f"Expected one of {list(_common_steps)}."
        ) from None


def parse_steps(
    raw_steps: list[str | dict[str, dict[str, str]]]
) -> Chain[Path, list[Atoms]]:
    steps: list[Step] = []

    for step in raw_steps:
        if isinstance(step, str):
            steps.append(get_step_type(step)())
        else:
            name, kwargs = list(step.items())[0]
            steps.append(get_step_type(name)(**kwargs))

    return Chain(steps)


def default_processing(file_name: str) -> Chain[Path, list[Atoms]]:
    default = f"""\
- SelectFile:
    file: "{file_name}"
- ReadASE
"""
    return parse_steps(yaml.safe_load(default))


@register_step
class UnZip(Step[Path, Path]):
    """
    UnZips an archive folder, and returns the path to the extracted folder.

    Parameters
    ----------
    file
        The name of the file to extract. If not provided, it is assumed that
        only a single archive file is present in the directory.

    Examples
    --------
    Unzip the downloaded file:

    .. code-block:: yaml

            processing:
            - UnZip

    Unzip a specific file:

    .. code-block:: yaml

        processing:
            - UnZip:
                file: "archive.zip"

    Behaviour:

    >>> UnZip()(Path("root"))
    Path("root/extracted")
    """

    def __init__(self, file: str | None = None):
        self.file = file

    def __call__(self, root: Path, progress: Progress | None = None) -> Path:
        if self.file is None:
            # get the first (and only) file in the directory
            extract_from = next(root.iterdir())
        else:
            extract_from = root / self.file
        extract_to = root / f"{extract_from.name}-extracted"

        if progress is not None:
            context = progress.new_task(f"Extracting {extract_from.name}")
        else:
            context = contextlib.nullcontext()
        with context:
            shutil.unpack_archive(extract_from, extract_dir=extract_to)
        return extract_to

    def __repr__(self):
        return "UnZip()" if self.file is None else f"UnZip({self.file})"


@register_step
class SelectFile(Step[Path, Path]):
    """
    Alters the input path to point to a specific file.

    Parameters
    ----------
    file
        The name of the file to select.

    Examples
    --------
    Select a specific file:

    .. code-block:: yaml

        processing:
            - SelectFile:
                file: "file.txt"

    Behaviour:

    >>> SelectFile("file.txt")(Path("root"))
    Path("root/file.txt")
    """

    def __init__(self, file: str):
        self.file = file

    def __call__(self, root: Path, progress: Progress | None = None) -> Path:
        return root / self.file

    def __repr__(self):
        return f"SelectFile({self.file})"


@register_step
class ForEachFile(Step[Path, "list[Atoms]"]):
    """
    Apply a chain of processing steps to each file in a directory.

    Parameters
    ----------
    steps
        The steps to apply to each file.
    files
        The files to apply the steps to. If not provided, all files in the
        directory are processed.
    pattern
        A pattern to match files in the directory. If not provided, all files
        in the directory are processed.

    Examples
    --------
    Apply a chain of processing steps to all files in a directory:

    .. code-block:: yaml

        processing:
            - ForEachFile:
                steps:
                  - ReadASE: {}
                files:
                  - "file1.xyz"
                  - "file2.xyz"
    """

    def __init__(
        self,
        steps: list[str | dict[str, dict[str, str]]],
        files: list[str] | None = None,
        pattern: str | None = None,
    ):
        self.chain = parse_steps(steps)
        self.files = files
        self.pattern = pattern

    def __call__(
        self, root: Path, progress: Progress | None = None
    ) -> list[Atoms]:
        if self.files is not None:
            files = [root / file for file in self.files]
        elif self.pattern is not None:
            files = sorted(root.glob(self.pattern))
        else:
            files = sorted(root.iterdir())

        if progress is None:
            results = []
            for file in files:
                results.extend(self.chain(file))
            return results

        results = []
        with progress.new_task("Processing files", total=len(files)) as task:
            for file in files:
                # explicitly ignore the progress
                results.extend(self.chain(file, progress=None))
                task.update(advance=1)

        return results


@register_step
class ReadASE(Step[Path, "list[Atoms]"]):
    """
    Read the atoms from an ASE-readable file.

    Parameters
    ----------
    kwargs
        Additional keyword arguments to pass to :func:`ase.io.read`.

    Examples
    --------
    Read all atoms from an xyz file:

    .. code-block:: yaml

        processing:
            - SelectFile:
                file: "structures.xyz"
            - ReadASE

    Behaviour:

    >>> ReadASE()(Path("root/structures.xyz"))
    [Atoms(...), Atoms(...), ...]
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(
        self, file: Path, progress: Progress | None = None
    ) -> list[Atoms]:
        if progress is not None:
            with progress.new_task(f"Reading {file.name}"):
                return ase_read(file, index=":", **self.kwargs)  # type: ignore

        return ase_read(file, index=":", **self.kwargs)  # type: ignore

    def __repr__(self):
        return f"ReadASE({self.kwargs})"


@register_step
class Custom(Step[Path, "list[Atoms]"]):
    """
    Apply a custom processing function for a dataset.

    This should be implemented in
    :code:`load_atoms.database.processing.custom.<id>::process`, taking the
    path to the download directory as an argument, and returning a list of
    :code:`ase.Atoms` objects.

    Parameters
    ----------

    id
        The id of the custom processing function to apply.

    Examples
    --------

    Apply a custom processing function:

    .. code-block:: yaml

        processing:
            - Custom:
                id: "<id>"
    """

    def __init__(self, id: str):
        self.id = id
        try:
            self.func = importlib.import_module(
                f"load_atoms.database.processing.{self.id}"
            ).process
        except ModuleNotFoundError:
            raise ValueError(f"Unknown custom processing: {self.id}") from None

    def __call__(
        self, file: Path, progress: Progress | None = None
    ) -> list[Atoms]:
        return self.func(file)

    def __repr__(self):
        return f"Custom({self.id})"


@register_step
class Rename(Step["list[Atoms]", "list[Atoms]"]):
    """
    Rename fields on the atoms objects.

    """

    def __init__(self, **name_mapping):
        self.name_mapping = name_mapping

    def __call__(
        self, atoms: list[Atoms], progress: Progress | None = None
    ) -> list[Atoms]:
        for atom in atoms:
            for old, new in self.name_mapping.items():
                if old in atom.arrays:
                    atom.arrays[new] = atom.arrays.pop(old)
                elif old in atom.info:
                    atom.info[new] = atom.info.pop(old)
        return atoms



import xml.etree.ElementTree as ET
from datetime import datetime
from pathlib import Path
from xml.dom import minidom

import numpy as np
from ase import Atom, Atoms
from ase.data import covalent_radii
from ase.data.colors import jmol_colors
from ase.neighborlist import natural_cutoffs, neighbor_list
from IPython.core.display import HTML

BOND_RADIUS = 0.17
MIN_ATOM_RADIUS = BOND_RADIUS + 0.1


def view(atoms: Atoms, show_bonds: bool = False):
    """
    Generate an interactive visualisation of an :class:`ase.Atoms` object.

    Parameters
    ----------
    atoms:
        The atoms object to visualise.
    show_bonds:
        Whether to show bonds between atoms.

    Example
    -------
    Visualise a water molecule in a Jupyter notebook:

    .. code-block:: python
        :emphasize-lines: 5

        from ase.build import molecule
        from load_atoms import view

        ethanol = molecule("CH3CH2OH")
        view(ethanol, show_bonds=True)

    .. raw:: html
        :file: ../../../docs/source/_static/ethanol.html

    Save the visualisation to an :code:`html` file:

    .. code-block:: python
        :emphasize-lines: 3

        from pathlib import Path

        viz = view(ethanol, show_bonds=True)
        Path("ethanol.html").write_text(viz.data)
    """
    scene = x3d_scene(atoms, show_bonds, width="300px", height="300px")

    uid = unique_variable_name()
    js_file = Path(__file__).parent / "view.js"
    js = js_file.read_text().replace("uid", uid)

    x3d_script = (Path(__file__).parent / "x3d.script").read_text()

    return HTML(
        f"""\
<html>
    <body>
    {x3d_script}
    <div id="{uid}">
        {scene}
    </div>
    <script>
        {js}
    </script>
    </body>
</html>
"""
    )


def x3d_scene(
    structure: Atoms,
    show_bonds: bool = False,
    **css_style: str,
) -> str:
    style = {"width": "300px", "height": "300px"}
    style.update(css_style)
    style = " ".join(f'{k}="{v}";' for k, v in style.items())

    scene = x3d_atoms(structure.copy(), show_bonds=show_bonds)

    document = f"""\
<X3D {style}>
<param name="showProgress" value="false" />

<!--Inserting Generated X3D Scene-->
{pretty_print(scene)}
<!--End of Inserted Scene-->

</X3D>
"""
    return document


def x3d_atom(atom: Atom, scale: float = 1.0):
    """Represent an atom as a coloured sphere."""

    radius = max(covalent_radii[atom.number] * scale, MIN_ATOM_RADIUS)
    shape = element(
        "shape",
        children=(appearance_for(atom), element("sphere", radius=f"{radius}")),
    )

    x, y, z = atom.position
    return translate(shape, x, y, z)


def appearance_for(atom: Atom):
    r, g, b = jmol_colors[atom.number]
    return element(
        "appearance",
        child=element("material", diffuseColor=f"{r} {g} {b}"),
    )


def x3d_wireframe_box(box):
    """
    Wireframe representation of a box (3x3 array).

    To draw a box, spanned by vectors a, b and c, it is necessary to
    draw 4 faces, each of which is a parallelogram. The faces are:
    (start from) , (vectors spanning the face)
    1. (0), (a, b)
    2. (c), (a, b) # opposite face to 1.
    3. (0), (a, c)
    4. (b), (a, c) # opposite face to 3.
    """

    a, b, c = box
    faces = [
        wireframe_face(a, b),
        wireframe_face(a, b, origin=c),
        wireframe_face(a, c),
        wireframe_face(a, c, origin=b),
    ]
    return group(faces)


def wireframe_face(vec1, vec2, origin=(0, 0, 0)):
    """Wireframe representation of a face spanned by vec1 and vec2."""

    x1, y1, z1 = vec1
    x2, y2, z2 = vec2

    material = element(
        "material", diffuseColor="0 0 0"
    )  # TODO: make this work in both dark and light mode
    appearance = element("appearance", child=material)

    points = [
        (0, 0, 0),
        (x1, y1, z1),
        (x1 + x2, y1 + y2, z1 + z2),
        (x2, y2, z2),
        (0, 0, 0),
    ]
    points = " ".join(f"{x} {y} {z}" for x, y, z in points)

    coordinates = element("coordinate", point=points)
    lineset = element("lineset", vertexCount="5", child=coordinates)
    shape = element("shape", children=(appearance, lineset))

    x, y, z = origin
    return translate(shape, x, y, z)


def x3d_atoms(atoms: Atoms, show_bonds: bool = False):
    """Convert an atoms object into an x3d representation."""

    # first, ensure that the atoms are within the cell
    if atoms.pbc.any():
        atoms.wrap()
        show_cell = True
    else:
        atoms.center(vacuum=0.1)
        show_cell = False

    scale = 0.6 if show_bonds else 1.0

    atom_spheres = group([x3d_atom(atom, scale) for atom in atoms])  # type: ignore
    bonds = x3d_bonds(atoms, scale) if show_bonds else empty_element()
    wireframe = x3d_wireframe_box(atoms.cell) if show_cell else empty_element()
    cell = group((wireframe, atom_spheres, bonds))

    # we want the cell to be in the middle of the viewport
    # so that we can (a) see the whole cell and (b) rotate around the center
    # therefore we translate so that the center of the cell is at the origin
    cell_center = atoms.cell.sum(axis=0) / 2  # type: ignore
    cell = translate(cell, *(-cell_center))
    # rotate the cell about y (we grab this later to start spinning it)
    cell = rotate(cell, 0, 1, 0, 0.1, id="atoms-rotation")

    # we want the cell, and all atoms, to be visible
    # - sometimes atoms appear outside the cell
    # - sometimes atoms only take up a small part of the cell
    # location of the viewpoint therefore takes both of these into account:
    # the scene is centered on the cell, so we find the furthest point away
    # from the cell center, and use this to determine the
    # distance of the viewpoint
    points = np.vstack((atoms.positions, atoms.cell[:]))  # type: ignore
    max_xyz_extent = get_maximum_extent(points - cell_center)

    # put the camera 2.2x as far away as
    # the largest separation between two points in any of x, y or z
    max_dim = max(max_xyz_extent)
    pos = f"0, 0, {max_dim * 2.2}"

    # NB. viewpoint needs to contain an (empty) child to be valid x3d
    viewpoint = element("viewpoint", position=pos, child=empty_element())

    return element("scene", children=(viewpoint, cell), id="atoms-scene")


def empty_element():
    return element("group")


def x3d_bonds(atoms: Atoms, scale: float = 1.0):
    # set pbc off so that we don't try to draw bonds across periodic boundaries
    atoms = atoms.copy()
    atoms.pbc = False
    i, j = neighbor_list("ij", atoms, cutoff=natural_cutoffs(atoms, mult=1.2))  # type: ignore
    bond_index = list(zip(i.tolist(), j.tolist()))

    return group([x3d_bond(atoms[a], atoms[b], scale) for a, b in bond_index])  # type: ignore


def atom_size(atom: Atom, scale: float = 1.0):
    return max(covalent_radii[atom.number] * scale, MIN_ATOM_RADIUS)


def x3d_bond(a: Atom, b: Atom, scale: float = 1.0):
    # if same element: draw a single cylinder connecting the two
    if a.number == b.number:
        return cyclinder_between(a.position, b.position, appearance_for(a))

    # otherwise, draw two cylinders, one from each atom to a suitable
    # point in the middle, and coloured according the respective atoms
    dist = np.linalg.norm(a.position - b.position)
    r_a = atom_size(a, scale)
    r_b = atom_size(b, scale)
    frac = r_a / dist + (0.5 * (dist - r_a - r_b)) / dist
    halfway = a.position + frac * (b.position - a.position)

    a_to_center = cyclinder_between(a.position, halfway, appearance_for(a))
    center_to_b = cyclinder_between(halfway, b.position, appearance_for(b))

    return group([a_to_center, center_to_b])


def cyclinder_between(a, b, appearance):
    halfway = (a + b) / 2

    z = np.array([0, 1, 0])
    v = b - a
    d = np.linalg.norm(v)
    v = v / d
    angle = np.arccos(z.dot(v))
    axis = np.cross(z, v)

    cylinder = element(
        "shape",
        children=[
            appearance,
            element("cylinder", radius=f"{BOND_RADIUS}", height=f"{d}"),
        ],
    )
    return translate(rotate(cylinder, *axis, angle), *halfway)  # type: ignore


def element(name, child=None, children=None, **attributes) -> ET.Element:
    """Convenience function to make an XML element.

    If child is specified, it is appended to the element.
    If children is specified, they are appended to the element.
    You cannot specify both child and children."""

    # make sure we don't specify both child and children
    if child is not None:
        assert children is None, "Cannot specify both child and children"
        children = [child]
    else:
        children = children or []

    element = ET.Element(name, **attributes)
    for child in children:
        element.append(child)
    return element


def translate(thing, x, y, z):
    """Translate a x3d element by x, y, z."""
    return element("transform", translation=f"{x} {y} {z}", child=thing)


def rotate(thing, x, y, z, angle, **attributes):
    return element(
        "transform",
        rotation=f"{x}, {y}, {z}, {angle}",
        child=thing,
        **attributes,
    )


def group(things):
    """Group a (list of) x3d elements."""
    return element("group", children=things)


def pretty_print(element: ET.Element, indent: int = 2):
    """Pretty print an XML element."""

    byte_string = ET.tostring(element, "utf-8")
    parsed = minidom.parseString(byte_string)
    prettied = parsed.toprettyxml(indent=" " * indent)
    # remove first line - contains an extra, un-needed xml declaration
    lines = prettied.splitlines()[1:]
    return "\n".join(lines)


def get_maximum_extent(xyz):
    """Get the maximum extent of an array of 3d set of points."""

    return np.max(xyz, axis=0) - np.min(xyz, axis=0)


def unique_variable_name():
    now = str(datetime.now().timestamp())
    now = now.replace(".", "")
    return "".join(chr(97 + int(digit)) for digit in now)



"""
The :code:`backend` module is responsible for down/loading datasets by name, 
storing them locally, and serving them to :code:`load-atoms` via the
:func:`~load_atoms.load_dataset` function.
"""

from __future__ import annotations

import pickle
import shutil
import warnings
from pathlib import Path

from ase import Atoms

from load_atoms.database.database_entry import DatabaseEntry, valid_licenses
from load_atoms.database.internet import download
from load_atoms.progress import Progress
from load_atoms.utils import (
    UnknownDatasetException,
    frontend_url,
    matches_checksum,
)

_REMOVE_TEMP = True


def load_structures(name: str, root: Path) -> tuple[list[Atoms], DatabaseEntry]:
    """
    Load the structures comprising the dataset with the given id from the
    given path.

    If these structures are not currently present, download and process
    them first.

    Parameters
    ----------
    name
        The id of the dataset to load.
    root
        The root folder to save the structures to.
    """
    with Progress(f"[bold]{name}") as progress:
        result = _load_structures(name, root, progress)
        progress._live.refresh()
    return result


def _load_structures(name, root, progress):
    entry_path = root / name / f"{name}.yaml"
    structures_path = root / name / f"{name}.pkl"
    temp_path = root / name / "temp"

    entry_path.parent.mkdir(parents=True, exist_ok=True)

    # we first need to get the DatabaseEntry for the dataset
    if not (entry_path).exists():
        try:
            download(
                DatabaseEntry.remote_url_for(name),
                entry_path,
                Progress("", transient=True),
            )
        except Exception as e:
            raise UnknownDatasetException(name) from e

    entry = DatabaseEntry.from_yaml_file(entry_path)

    # try to load the structures from disk
    if structures_path.exists():
        with progress.new_task("Reading from disk"), open(
            structures_path, "rb"
        ) as f:
            structures = pickle.load(f)

        log_usage_information(entry, progress)

        return structures, entry

    # else, download and process the dataset

    # 1. download the dataset files
    temp_path.mkdir(parents=True, exist_ok=True)

    # the following lines add no performance overhead, but are useful for
    # when debugging the adding of new datasets
    existing_files: list[str] = [f.name for f in temp_path.iterdir()]
    for file in entry.files:
        if file.name in existing_files:
            continue

        download(
            file.url,
            temp_path / file.name,
            progress,
        )

    # 2. validate the downloaded files
    for file_info in entry.files:
        file = temp_path / file_info.name
        if not matches_checksum(file, file_info.hash):
            warnings.warn(
                f"Checksum of {file_info.name} does not match the "
                "expected value. This means that the downloaded dataset "
                "may be corrupted.",
                stacklevel=2,
            )

    # 3. process the downloaded files into structures
    structures = entry.processing(temp_path, progress)
    # remove annoying default calculators
    for s in structures:
        s.calc = None

    # 4. save the structures to disk
    with progress.new_task("Caching to disk"), open(structures_path, "wb") as f:
        pickle.dump(structures, f)

    # 5. clean up the download directory
    # if debugging, comment out this line to inspect the downloaded files
    # rather than deleting them
    if _REMOVE_TEMP:
        with progress.new_task("Cleaning up", transient=True):
            shutil.rmtree(temp_path)

    log_usage_information(entry, progress)

    return structures, entry


def log_usage_information(info: DatabaseEntry, progress: Progress):
    progress.log_below("\n")

    name = f"[bold]{info.name}[/bold]"
    if info.license is not None:
        style = f"dodger_blue2 link={valid_licenses[info.license]} underline"
        progress.log_below(
            f"The {name} dataset is covered by the "
            f"[{style}]{info.license}[/] license."
        )
    if info.citation is not None:
        progress.log_below(
            f"Please cite the {name} dataset " "if you use it in your work."
        )
    progress.log_below(f"For more information about the {name} dataset, visit:")
    url = frontend_url(info)
    url_style = f"dodger_blue2 underline link={url}"
    progress.log_below(f"[{url_style}]load-atoms/{info.name}")
